// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

package org.columba.composer;

import java.io.StringReader;
import java.util.Date;
import java.util.Vector;

import org.columba.main.*;
import org.columba.gui.composer.*;
import org.columba.gui.statusbar.*;
import org.columba.message.*;
import org.columba.config.*;
import org.columba.util.*;

public class MessageComposer
{
	private ComposerInterface composerInterface;
	private String bound;
	private int accountUid;

	public MessageComposer(ComposerInterface iface)
	{
		composerInterface = iface;
	}

	private ColumbaHeader initHeader(IdentityItem identity)
	{
		ColumbaHeader header = new ColumbaHeader();

		// RFC822 - Header

		if (composerInterface.composerHeader.getToSize() > 0)
			header.set("To", composerInterface.composerHeader.getToList());

		if (composerInterface.composerHeader.getCcSize() > 0)
			header.set("Cc", composerInterface.composerHeader.getCcList());

		header.set("Subject", composerInterface.composerEditor.getSubject());
		header.set("From", composerInterface.composerEditor.getFromAddress());
		header.set("X-Priority", composerInterface.composerEditor.getPriority());

		String priority = (String) header.get("X-Priority");
		if (priority != null)
		{
			header.set("columba.priority", new Integer(priority));
		}
		else
		{
			header.set("columba.priority", new Integer(3));
		}

		header.set("Mime-Version", "1.0");

		String organisation = identity.getOrganisation();
		if (organisation.length() > 0)
			header.set("Organisation", organisation);

		// reply-to
		String replyAddress = identity.getReplyAddress();
		if (replyAddress.length() > 0)
			header.set("Reply-To", replyAddress);

		String messageID = MessageIDGenerator.generate();
		header.set("Message-Id", messageID);

		String inreply =
			composerInterface.composerHeader.getAdditionalHeaderItem("In-Reply-To");
		if (inreply != null)
			header.set("In-Reply-To", inreply);

		String references =
			composerInterface.composerHeader.getAdditionalHeaderItem("References");
		if (references != null)
			header.set("References", references);

		header.set("X-Mailer", "Columba v" + org.columba.main.MainInterface.version);

		// shortFrom
		String shortFrom = (String) header.get("From");
		if (shortFrom != null)
		{
			if (shortFrom.indexOf("<") != -1)
			{
				shortFrom = shortFrom.substring(0, shortFrom.indexOf("<"));
				if (shortFrom.length() > 0)
				{
					if (shortFrom.startsWith("\""))
						shortFrom = shortFrom.substring(1, shortFrom.length() - 1);
					if (shortFrom.endsWith("\""))
						shortFrom = shortFrom.substring(0, shortFrom.length() - 1);
				}

			}
			else
				shortFrom = shortFrom;

			header.set("columba.from", shortFrom);
		}
		else
		{
			header.set("columba.from", new String(""));
		}

		// date
		Date date = new Date();
		header.set("columba.date", date);
		header.set("Date", RFC822Date.toRFC822String(date));

		return header;
	}

	private MimePart composeBody()
	{
		MimePart bodyPart = new MimePart();
		// Init Mime-Header with Default-Values (text/plain)

		String body = composerInterface.composerEditor.getBody();

		String signature = composerInterface.composerEditor.getSignature();

		if (signature != null)
		{
			body = body + "\n\n" + signature;
		}

		if (body.length() == 0)
		{
			return null;
		}

		bodyPart.setBody(body);

		return bodyPart;
	}

	public SendableMessage compose(int i)
	{
		this.accountUid = i;

		SendableMessage message = new SendableMessage();

		StringBuffer composedMessage = new StringBuffer();

		MimeHeader bodyHeader;

		int uid = composerInterface.composerEditor.getAccountUid();

		AccountItem item = composerInterface.config.getAccountList().uidGet(uid);
		IdentityItem identity = item.getIdentityItem();

		ColumbaHeader header = initHeader(identity);
		MimePart body = composeBody();
		MimePartCollection mimeCollection =
			composerInterface.composerAttachment.getAttachmentCollection();
		String composedBody;

		// Is this a MultipartMessage ?

		if (body != null)
			mimeCollection.insert(0,body);

		if (mimeCollection.count() == 0)
		{ // Add dummyline as body
			body = new MimePart();
			body.setBody(new String("\n"));
			mimeCollection.add(body);
		}

		if (mimeCollection.count() == 1)
		{

			// No Multipart Message

			composedBody = mimeCollection.get(0).composeToString();
		}
		else
		{

			// Mulitpart Message

			header.set("columba.attachment", new Boolean(true));

			composedBody = createMimeMultipart(mimeCollection);
		}

		composedMessage.append(header.getHeader());
		//	composedMessage.append( "\n" );
		composedMessage.append(composedBody);
		composedMessage.append("\n");

		message.setRawString(composedMessage.toString());

		//	System.out.println( "<-------Composed Message\n"+composedMessage+"------------------>\n");

		// size
		int size = composedMessage.length() / 1024;
		header.set("columba.size", new Integer(size));

		message.setHeader(header);

		message.setAccountUid(uid);

		// Body

		//        message.setBodyText( composedBody );

		// Attachments

		message.setMimePartCollection(mimeCollection);

		boolean sign = composerInterface.composerEditor.getSign();
		boolean encrypt = composerInterface.composerEditor.getEncrypt();

		System.out.println("sign: " + sign);
		System.out.println("-------------------->encrypted: " + encrypt);

		signEncryptMessage(sign, encrypt, message, message.getRawString());

		return message;
	}

	protected void signEncryptMessage(
		boolean sign,
		boolean encrypt,
		SendableMessage message,
		String body)
	{
		if (encrypt == true)
		{
			System.out.println("creating encrypted message:");

			if (bound == null)
				bound = createBoundary();

			Vector recipients = composerInterface.composerHeader.getRCPTVector();

			AccountItem item = composerInterface.config.getAccountList().uidGet(accountUid);
			PGPItem pgpItem = item.getPGPItem();

			String encryptedMessage =
				composerInterface.pgpController.encrypt(body, sign, recipients, pgpItem);

			MimePartCollection col = new MimePartCollection();
			MimePart part = new MimePart();
			part.getHeader().setContentTransferEncoding("7bit");
			part.getHeader().contentType = "application";
			part.getHeader().contentSubtype = "pgp-encrypted";
			part.setBody("Version: 1");
			col.add(part);
			part = new MimePart();
			part.getHeader().setContentTransferEncoding("7bit");
			part.getHeader().contentType = "application";
			part.getHeader().contentSubtype = "octet-stream";
			part.setBody(encryptedMessage);
			col.add(part);

			message.setMimePartCollection(col);

			StringBuffer contents = new StringBuffer();
			contents.append(message.getHeader().getHeader());

			contents.append(createEncryptedMimePart(col));

			message.setRawString(contents.toString());

			//System.out.println("encrypted message:\n"+ contents );

		}
		/*
		else if ( sign == true )
		{
		}
		*/

	}

	protected String[] divideMessage(String input)
	{
		String[] output = new String[2];
		int emptyLinePos;

		if (input.charAt(0) == '\n')
		{
			output[1] = input;
			return output;
		}

		emptyLinePos = input.indexOf("\n\n");

		if (input.indexOf("\n\n") != -1)
		{
			output[0] = input.substring(0, emptyLinePos + 1);
			output[1] = input.substring(emptyLinePos + 2);

			return output;
		}

		return null;
	}

	private String createBoundary()
	{
		StringBuffer bound = new StringBuffer();

		bound.append("-----");
		bound.append(System.currentTimeMillis());
		bound.append(composerInterface.composerEditor.getFromAddress());

		return bound.toString();
	}

	private String createMimeMultipart(MimePartCollection input)
	{
		StringBuffer output = new StringBuffer();
		bound = createBoundary();
		MimeHeader multiHeader = new MimeHeader("multipart", "mixed");
		multiHeader.contentParameter.put("boundary", new String("\"" + bound + "\""));

		output.append(multiHeader.getHeader());
		output.append("\n\tThis is a Mime Multipart Message sent with Columba\n");

		int count = input.count();

		for (int i = 0; i < count; i++)
		{
			output.append("\n--" + bound + "\n");
			output.append(input.get(i).composeToString());
		}

		output.append("\n--" + bound + "--");

		return output.toString();
	}

	private String createEncryptedMimePart(MimePartCollection input)
	{
		if (input.count() > 1)
			return createEncryptedMimeMultipart(input);

		StringBuffer output = new StringBuffer();
		MimePart actPart = input.get(0);

		return actPart.composeToString();
	}

	private String createEncryptedMimeMultipart(MimePartCollection input)
	{

		StringBuffer output = new StringBuffer();
		MimeHeader multiHeader = new MimeHeader("multipart", "encrypted");
		multiHeader.contentParameter.put("protocol", "\"application/pgp-encrypted\"");
		bound = bound.substring(2, bound.length());
		multiHeader.contentParameter.put("boundary", new String("\"" + bound + "\""));

		output.append(multiHeader.getHeader());
		output.append("\n\tThis is a Mime Multipart Message sent with Columba\n");

		int count = input.count();

		for (int i = 0; i < count; i++)
		{
			output.append("\n--" + bound + "\n");
			output.append(input.get(i).composeToString());
		}

		output.append("\n--" + bound + "--");

		return output.toString();
	}

}