// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/** Document the purpose of this class.
 *
 * @version 1.0
 * @author Timo Stich
 */

package org.columba.parser;

import java.io.*;
import org.columba.coder.*;
import org.columba.message.*;

public class Rfc822Parser extends AbstractParser
{

	// Definition of Constants

	private static final int RETURN_PATH = 0;
	private static final int RECEIVED = 1;
	private static final int FROM = 2;
	private static final int SENDER = 3;
	private static final int REPLY_TO = 4;
	private static final int RESENT_FROM = 5;
	private static final int RESENT_SENDER = 6;
	private static final int RESENT_REPLY_TO = 7;
	private static final int DATE = 8;
	private static final int RESENT_DATE = 9;
	private static final int TO = 10;
	private static final int RESENT_TO = 11;
	private static final int CC = 12;
	private static final int RESENT_CC = 13;
	private static final int BCC = 14;
	private static final int RESENT_BCC = 15;
	private static final int MESSAGE_ID = 16;
	private static final int RESENT_MESSAGE_ID = 17;
	private static final int IN_REPLY_TO = 18;
	private static final int REFERENCES = 19;
	private static final int KEYWORDS = 20;
	private static final int SUBJECT = 21;
	private static final int COMMENTS = 22;
	private static final int ENCRYPTED = 23;
	private static final int Mime_VERSION = 24;
	private static final int X_ADDITIONAL = 25;

	// Global Variables
	private String[] input;
	private EncodedWordDecoder decoder;
	private ColumbaHeader parsedHeader;
	private boolean mime;
	private String mimeVer;

	// Constructor

	// Public Methods
	public Rfc822Parser()
	{

	}

	// the boolean parseHeader argument:
	// if true the header gets parsed, otherwise not
	//
	// action can be 0 or 1:
	// 0 means: just parse bodystructure and headers
	// 1 means: if encrypted or signed use pgp to read message
	public Message parse(
		String input,
		boolean parseHeader,
		ColumbaHeader header,
		int action)
	{

		//System.out.println("------------------------------Parsing RFC822");

		Message output = new Message();
		MimePartCollection mimeParts = new MimePartCollection();
		this.input = divideMessage(input);

		decoder = new EncodedWordDecoder();

		if (parseHeader)
		{
			ColumbaHeader tempHeader = parseHeaderString(this.input[0]);
			//System.out.println( tempHeader.get( "mime-version" ));
			tempHeader.copyColumbaKeys(header);
			output.setHeader(tempHeader);
		}
		else
		{
			output.setHeader(header);
		}

		String mimeValue = (String) output.getHeader().get("mime-version");
		//System.out.println( mimeValue );

		if (mimeValue != null)
		{
			mime = true;
			/*
			  Commented for more robustness with MIME-Version
			
			        if ( mimeValue.equals("1.0") )
			{
			    mimeVer = new String("1.0");
			    mime=true;
			}
			        else
			{
			
			    mime=false;
			}
			*/
		}
		else
		{
			//System.out.println("---------------------------NOT Parsing MIME");
			mime = false;
		}

		if (mime)
		{
			MimeParser mimeParser = new MimeParser(header, action);

			//System.out.println("------------------------------Parsing MIME");

			mimeParts = mimeParser.parse(input);
		}
		else
		{
			MimeHeader helpHeader = new MimeHeader();

			mimeParts.add(new MimePart(helpHeader, this.input[1], null));
		}

		output.setMimePartCollection(mimeParts);

		return output;
	}

	// Private Methods

	// Method that returns the Code for the Rfc822 Header-Entity

	private int getCode(String par)
	{
		if (par.equals("Return-Path"))
			return RETURN_PATH;
		if (par.equals("Received"))
			return RECEIVED;
		if (par.equals("From"))
			return FROM;
		if (par.equals("Sender"))
			return SENDER;
		if (par.equals("Reply-To"))
			return REPLY_TO;
		if (par.equals("Resent-From"))
			return RESENT_FROM;
		if (par.equals("Resent-Sender"))
			return RESENT_SENDER;
		if (par.equals("Resent-Reply-To"))
			return RESENT_REPLY_TO;
		if (par.equals("Date"))
			return DATE;
		if (par.equals("Resent-Date"))
			return RESENT_DATE;
		if (par.equals("To"))
			return TO;
		if (par.equals("Resent-To"))
			return RESENT_TO;
		if (par.equals("cc"))
			return CC;
		if (par.equals("Resent-cc"))
			return RESENT_CC;
		if (par.equals("bcc"))
			return BCC;
		if (par.equals("Resent-bcc"))
			return RESENT_BCC;
		if (par.equals("Message-ID"))
			return MESSAGE_ID;
		if (par.equals("Resent-Message-ID"))
			return RESENT_MESSAGE_ID;
		if (par.equals("In-Reply-To"))
			return IN_REPLY_TO;
		if (par.equals("References"))
			return REFERENCES;
		if (par.equals("Keywords"))
			return KEYWORDS;
		if (par.equals("Subject"))
			return SUBJECT;
		if (par.equals("Comments"))
			return COMMENTS;
		if (par.equals("Encrypted"))
			return ENCRYPTED;

		if (par.equals("Mime-Version"))
			return Mime_VERSION;

		if (par.startsWith("x"))
			return X_ADDITIONAL;
		if (par.startsWith("X"))
			return X_ADDITIONAL;

		return -1;
	}

	// This Method is the Heart of the class as it parses the Header

	public ColumbaHeader parseHeader(String message)
	{
		if (message == null)
		{

		}
		String[] divided = divideMessage(message);
		return parseHeaderString(divided[0]);
	}

	private ColumbaHeader parseHeaderString(String input) //, ColumbaHeader output )
	{
		String header;
		ColumbaHeader output = new ColumbaHeader();
		String line;
		StringBuffer received;

		//        if ( output == null ) output = new ColumbaHeader();

		// Check if Parameter is separated Header or whole Message
		if (decoder == null)
			decoder = new EncodedWordDecoder();

		HeaderTokenizer tokenizer = new HeaderTokenizer(input);

		line = tokenizer.nextLine();

		while (line != null)
		{
			if (line.indexOf(':') != -1)
			{

				String key = line.substring(0, line.indexOf(':'));
				key = key.toLowerCase();

				output.set(key, decoder.decode(line.substring(line.indexOf(':') + 1).trim()));
			}
			line = tokenizer.nextLine();
		}

		return output;
	}

}