// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


package org.columba.protocol;

import java.io.*;
import java.net.*;
import java.security.*;
import java.util.*;
import javax.swing.*;


import org.columba.util.*;
import org.columba.parser.*;

public class Imap4 {
	public static final int DEFAULT_PORT = 143;

	public static final int ANSWER_OK = 0;
	public static final int ANSWER_MORE = 1;
	public static final int ANSWER_CONTINUATION = 2;
	public static final int ANSWER_NO = 10;
	public static final int ANSWER_BAD = 11;
	public static final int UNKNOWN_ANSWER = 12;

	public static final int STATE_NONAUTHENTICATE = 0;
	public static final int STATE_AUTHENTICATE = 1;
	public static final int STATE_SELECTED = 2;

	private Socket socket;
	private BufferedReader in;
	private PrintWriter out;

	public String answer;

	private String userId = new String("A");
	private int id = 0;
	private String idString;

	private Vector capabilities;
	private int state; // non-authenticate=0, authenticate=1, selected=2,

	private String selectedFolderPath;
	private String result;

	public Imap4() {
		capabilities = new Vector();
		setState(STATE_NONAUTHENTICATE);
	}

	public void setState(int i) {
		state = i;
	}

	public int getState() {
		return state;
	}

	public String getResult() {
		return result;
	}

	public String getSelectedFolderPath() {
		return selectedFolderPath;
	}

	public void setSelectedFolderPath(String s) {
		selectedFolderPath = s;
	}

	public boolean openPort(String host) throws IOException {
		return openPort(host, DEFAULT_PORT);
	}

	public boolean openPort(String host, int port) throws IOException {
		socket = new Socket(host, port);
		in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
		out = new PrintWriter(new OutputStreamWriter(socket.getOutputStream()));

		answer = in.readLine();

		if (answer.startsWith("*"))
			return true;
		else
			return false;
	}

	protected String generateIdentifier() {
		id++;
		idString = new String(userId + new Integer(id).toString());
		return idString;
	}

	public void sendString(String s) throws IOException {
		String str = generateIdentifier() + " " + s + "\r\n";
		out.print(str);
		out.flush();

		// this is only for debugging
		//System.out.println("CLIENT: " + str.substring(0, str.length() - 2));
	}

	// send without creating a new id
	// this is only for message appending, where we have to send
	// a whole message to the server, without the generation
	// of a general client command
	public void send(String s) throws IOException {
		String str = s + "\r\n";
		out.print(str);
		out.flush();

		// this is only for debugging
		//System.out.println("CLIENT: " + str.substring(0, str.length() - 2));
	}

	public int getAnswer() throws IOException {
		answer = in.readLine();

		// this is only for debugging
		//System.out.println("SERVER: " + answer);

		if (answer.startsWith("*")) {
			return ANSWER_MORE;
		} else if (answer.startsWith("+")) {
			return ANSWER_CONTINUATION;
		} else if (answer.startsWith("BAD")) {
			return ANSWER_BAD;
		} else if (answer.indexOf("NO") != -1) {
			return ANSWER_NO;
		} 
		else if (answer.startsWith("A")) {
			int index = answer.indexOf(" ");
			String substring = answer.substring(index + 1, answer.length());

			if (substring.startsWith("OK")) {
				return ANSWER_OK;
			} else {
				return UNKNOWN_ANSWER;
			}
		} else {
			return UNKNOWN_ANSWER;
		}
	}

	public boolean processAnswer() throws IOException {
		result = new String();

		int i = getAnswer();

		if (i == ANSWER_MORE) {
			StringBuffer buf = new StringBuffer();

			while (i == ANSWER_MORE) {
				buf.append(answer + "\n");
				i = getAnswer();
			}
			if (i == ANSWER_OK) {
				buf.append(answer + "\n");
				result = buf.toString();

				return true;
			}

			return false;
		} else if (i == ANSWER_OK) {
			result = answer;
			return true;
		} else if (i == ANSWER_NO) {
			result = answer;
			return false;
		} else if (i == ANSWER_BAD) {
			result = answer;
			return false;
		} else
			return false;
	}

	/**************** any state commands ***********************/

	// this is not used right now
	public boolean capability() throws IOException {
		sendString("CAPABILITY");

		if (getAnswer() == ANSWER_MORE) {
			String token = answer;
			int index = token.toLowerCase().indexOf("imap4rev1");

			token =
				token.substring(index + new String("imap4rev1").length(), token.length());

			StringTokenizer tok = new StringTokenizer(token);
			while (tok.hasMoreElements()) {
				capabilities.add(tok.nextElement());
			}

			if (getAnswer() == ANSWER_OK)
				return true;
		}

		return false;
	}

	/*
	// do we really have to use noops ?
	public boolean noop() throws IOException
	{
	    if ( !in.ready() )
	    {
	        sendString("NOOP");
	        return processOneLineAnswer();
	    }
	    else return false;
	}
	*/

	public boolean logout() throws IOException {
		if (socket == null)
			return true;

		sendString("LOGOUT");

		return processAnswer();
	}

	public boolean login(String user, String password) throws IOException {
		// reset everything
		setState(STATE_NONAUTHENTICATE);
		setSelectedFolderPath("");

		sendString("LOGIN " + user + " " + password);

		boolean b = processAnswer();

		if (b == true) {
			// authenticate success
			setState(STATE_AUTHENTICATE);
		} else {
			setState(STATE_NONAUTHENTICATE);
		}

		return b;

	}

	/************************ authenticate state commands **************************/

	// we cannot use the processAnswer-method, because
	// we want control over the progressbar
	//
	// the method assumes the the maximum progressbar value
	// is already set correctly
	public String getUIDL(SwingWorker worker) throws IOException {
		sendString("FETCH 1:* (UID)");

		StringBuffer buffer = new StringBuffer();
		int i = getAnswer();

		while (i != Imap4.ANSWER_OK) {
			if (worker != null)
				worker.incProgressBarValue(1);

			buffer.append(answer + "\n");
			i = getAnswer();
		}

		return buffer.toString();
	}
	
	public String getUIDL(String messageSet, SwingWorker worker) throws IOException {
		sendString("FETCH "+messageSet+" (UID)");

		StringBuffer buffer = new StringBuffer();
		int i = getAnswer();

		while (i != Imap4.ANSWER_OK) {
			if (worker != null)
				worker.incProgressBarValue(1);

			buffer.append(answer + "\n");
			i = getAnswer();
		}

		return buffer.toString();
	}

	public boolean select(String messagebox) throws IOException {

		sendString("SELECT \"" + messagebox + "\"");

		boolean b = processAnswer();

		if (b == true) {
			// select success
			setState(STATE_SELECTED);
			setSelectedFolderPath(messagebox);
		} else {
			setState(STATE_AUTHENTICATE);
		}

		return b;

	}

	public boolean subscribe(String path) throws IOException {
		sendString("SUBSCRIBE \"" + path + "\"");

		boolean b = processAnswer();

		return b;

	}

	public boolean unsubscribe(String path) throws IOException {
		sendString("UNSUBSCRIBE \"" + path + "\"");

		boolean b = processAnswer();

		return b;

	}

	public boolean internCopy(String messageSet, String messageboxName)
		throws IOException {
		sendString(" UID COPY " + messageSet + " " + messageboxName);

		boolean b = processAnswer();

		return b;
	}

	
	public boolean flsub(String referenceName)
		throws IOException {

		sendString("LSUB \"\" \""+referenceName+"\"");
		
		

		
		int i = getAnswer();
		//System.out.println("answer: "+i);
		//System.out.println("answer: "+answer);

		if (i == ANSWER_MORE) {
			//System.out.println("ANSWER MORE");

			StringBuffer buf = new StringBuffer();

			while (i == ANSWER_MORE) {
				//System.out.println("ANSWER MORE");
				buf.append(answer + "\n");
				i = getAnswer();
				//System.out.println("answer: "+ buf);

			}
			if (i == ANSWER_OK) {
				//System.out.println("ANSWER OK");
				//buf.append( answer+"\n");
				result = buf.toString();
				//System.out.println("lsub-result: "+ result);
				return true;
			}

			return false;
		} else if (i == ANSWER_OK) {
			result = "";
			//result = answer;
			//System.out.println("ANSWER OK");
			//System.out.println("lsub answer ok");
			return true;
		} else if (i == ANSWER_NO) {
			//System.out.println("lsub failure: can't list that reference or name");
			return false;
		} else if (i == ANSWER_BAD) {
			//System.out.println("lsub answer basd");
			return false;
		} else {
			//System.out.println("lsub failure else");
			return false;
		}
		
		

	}

	public boolean flist(String referenceName, String messageboxName)
		throws IOException {
		sendString("LIST \"\" \"" + referenceName + "\"");

		//boolean b = processAnswer();

		//return b;
		
		result = new String();

		int i = getAnswer();

		if (i == ANSWER_MORE) {
			StringBuffer buf = new StringBuffer();

			while (i == ANSWER_MORE) {
				buf.append(answer + "\n");
				i = getAnswer();
			}
			if (i == ANSWER_OK) {
				//buf.append(answer + "\n");
				result = buf.toString();

				return true;
			}

			return false;
		} else if (i == ANSWER_OK) {
			result = "";
			return true;
		} else if (i == ANSWER_NO) {
			result = answer;
			return false;
		} else if (i == ANSWER_BAD) {
			result = answer;
			return false;
		} else
			return false;

	}
	
	
	public boolean create(String messageboxName) throws IOException {
		sendString("CREATE \"" + messageboxName + "\"");

		boolean b = processAnswer();

		return b;

	}

	public boolean delete(String messageboxName) throws IOException {
		sendString("DELETE \"" + messageboxName + "\"");

		boolean b = processAnswer();

		return b;

	}

	public boolean rename(String messageboxName, String newName)
		throws IOException {
		sendString("RENAME \"" + messageboxName + "\" \"" + newName + "\"");

		boolean b = processAnswer();

		return b;

	}

	public boolean append(String messagebox, String message) throws IOException {
		sendString("APPEND \"" + messagebox + "\" (\\SEEN) {" + message.length() + "}");
		int i = getAnswer();

		if (i == ANSWER_CONTINUATION) {
			
			send(message);
			
		}

		boolean b = processAnswer();

		return b;

	}

	/******************************** selected state commands *************************/

	public boolean close() throws IOException {
		sendString("CLOSE");

		boolean b = processAnswer();

		return b;

	}

	public boolean fetch(String messageSet, String dataItem) throws IOException {

		sendString("FETCH " + messageSet + " " + dataItem);

		boolean b = processAnswer();

		return b;

	}

	public boolean store(String messageSet, String dataItem, String dataItemValue)
		throws IOException {
		sendString(
			"UID STORE " + " " + messageSet + " " + dataItem + " " + dataItemValue);

		boolean b = processAnswer();

		return b;

	}

	public boolean expunge() throws IOException {
		sendString("EXPUNGE");

		boolean b = processAnswer();

		return b;

	}

	public boolean uid(String commandName, String commandArguments)
		throws IOException {
		sendString("UID" + " " + commandName + " " + commandArguments);

		boolean b = processAnswer();

		return b;

	}

	public String fetchMessageSource(Object uid, SwingWorker worker)
		throws Exception {
		sendString("UID FETCH " + uid + " BODY[]");

		int i = 0;
		StringBuffer buffer = new StringBuffer();
		i = getAnswer();

		int size = parseTransferLength(answer);

		if (size != -1) {
			if (worker != null) {
				worker.setProgressBarMaximum(size);
				worker.setProgressBarValue(0);
			}
		}

		int actualMessageSize = 0;
		while (i != Imap4.ANSWER_OK) {
			if (i == Imap4.ANSWER_MORE) {

				//cache.append( imap.answer+"\n");

				//Flags flags = changeFlags( cache.toString() );

			} else {
				buffer.append(answer + "\n");

				actualMessageSize += (getAnswerLength(answer + "\r\n"));
				//System.out.println("actual size:" + actualMessageSize);
				if (worker != null)
					worker.setProgressBarValue(actualMessageSize);
			}
			i = getAnswer();
		}

		// remove the ending ")"
		if (buffer.length() > 2)
			buffer.setLength(buffer.length() - 2);

		return buffer.toString();
	}

	public String fetchMimePart(Object uid, int number, SwingWorker worker)
		throws Exception {

		int i = 0;
		StringBuffer buffer = new StringBuffer();
		// fetch MimePart
		sendString("UID FETCH " + uid + " (BODY[" + (number + 1) + "])");

		buffer = new StringBuffer();
		i = getAnswer();

		int size = parseTransferLength(answer);

		if (size != -1) {
			if (worker != null) {
				worker.setProgressBarMaximum(size);
				worker.setProgressBarValue(0);
			}
		}

		int actualMessageSize = 0;
		while (i != Imap4.ANSWER_OK) {
			if (i == Imap4.ANSWER_MORE) {
				//cache.append( imap.answer+"\n");
			} else {
				buffer.append(answer + "\n");
				actualMessageSize += (getAnswerLength(answer + "\r\n"));
				//System.out.println("actual size:" + actualMessageSize);
				if (worker != null)
					worker.setProgressBarValue(actualMessageSize);
			}

			i = getAnswer();
		}

		// remove the ending ")"
		if (buffer.length() > 2)
			buffer.setLength(buffer.length() - 2);

		return buffer.toString();
	}

	protected int parseTransferLength(String s) {
		int size = -1;
		int leftsizepos = answer.indexOf("{");
		int rightsizepos = answer.indexOf("}");
		if ((leftsizepos != -1) && (rightsizepos != -1)) {
			// we found a size we can use to set the progressbar appropriatly
			leftsizepos += 1;
			String sizestr = answer.substring(leftsizepos, rightsizepos);
			//System.out.println("sizestring:" + sizestr);

			try {
				size = Integer.parseInt(sizestr);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		return size;
	}

	protected int getAnswerLength(String s) {
		return s.length();
	}


	public Vector searchHeader(
		String messageSet,
		String headerField,
		String pattern,
		boolean condition)
		throws Exception
	{
		Vector result = new Vector();

		if (messageSet.length() != 0)
		{
			if ( condition == true )
				sendString("SEARCH " + messageSet + " HEADER " + headerField + " " + pattern);
			else
				sendString("SEARCH " + messageSet + " HEADER NOT " + headerField + " " + pattern);
		}
		else
		{
			if ( condition == true )
				sendString("SEARCH HEADER " + headerField + " " + pattern);
			else
				sendString("SEARCH HEADER NOT " + headerField + " " + pattern);
			
		}

		int i = 0;
		StringBuffer buffer = new StringBuffer();
		i = getAnswer();

		while (i != Imap4.ANSWER_OK)
		{
			if (i == Imap4.ANSWER_MORE)
			{
				buffer.append(answer + "\n");

				//cache.append( imap.answer+"\n");

				//Flags flags = changeFlags( cache.toString() );
			}
			else
			{

			}
			i = getAnswer();
		}

		System.out.println("buffer:\n" + buffer);
		
		result = Imap4Parser.parseSearch( buffer.toString() );

		return result;
	}
	
	
	public Vector searchBody(
		String messageSet,
		String pattern,
		boolean condition)
		throws Exception
	{
		Vector result = new Vector();

		if (messageSet.length() != 0)
		{
			if ( condition == true )
				sendString("SEARCH " + messageSet + " BODY"  + " " + pattern);
			else
				sendString("SEARCH " + messageSet + " BODY" + " NOT " + pattern);
		}
		else
		{
			if ( condition == true )
				sendString("SEARCH " + "BODY" + " " + pattern);
			else
				sendString("SEARCH "+ "NOT " + "BODY " + pattern);
			
		}

		int i = 0;
		StringBuffer buffer = new StringBuffer();
		i = getAnswer();

		while (i != Imap4.ANSWER_OK)
		{
			if (i == Imap4.ANSWER_MORE)
			{
				buffer.append(answer + "\n");

				//cache.append( imap.answer+"\n");

				//Flags flags = changeFlags( cache.toString() );
			}
			else
			{

			}
			i = getAnswer();
		}

		System.out.println("buffer:\n" + buffer);
		
		result = Imap4Parser.parseSearch( buffer.toString() );

		return result;
	}
	
	public Vector searchDate(
		String messageSet,
		String pattern,
		boolean condition)
		throws Exception
	{
		Vector result = new Vector();

		if (messageSet.length() != 0)
		{
			if ( condition == true )
				sendString("SEARCH " + messageSet + " SENTBEFORE"  + " " + pattern);
			else
				sendString("SEARCH " + messageSet + " SENTSINCE" + pattern);
		}
		else
		{
			if ( condition == true )
				sendString("SEARCH " + "SENTBEFORE" + " " + pattern);
			else
				sendString("SEARCH "+ "SENTSINCE " + pattern);
			
		}

		int i = 0;
		StringBuffer buffer = new StringBuffer();
		i = getAnswer();

		while (i != Imap4.ANSWER_OK)
		{
			if (i == Imap4.ANSWER_MORE)
			{
				buffer.append(answer + "\n");

				//cache.append( imap.answer+"\n");

				//Flags flags = changeFlags( cache.toString() );
			}
			else
			{

			}
			i = getAnswer();
		}

		System.out.println("buffer:\n" + buffer);
		
		result = Imap4Parser.parseSearch( buffer.toString() );

		return result;
	}
	
	public Vector searchSize(
		String messageSet,
		String pattern,
		boolean condition)
		throws Exception
	{
		Vector result = new Vector();

		int size = Integer.parseInt( pattern );
		size = size*1024;
		
		String sizeString = ( new Integer(size) ).toString();
		
		if (messageSet.length() != 0)
		{
			if ( condition == true )
				sendString("SEARCH " + messageSet + " SMALLER"  + " " + sizeString);
			else
				sendString("SEARCH " + messageSet + " LARGER" + sizeString);
		}
		else
		{
			if ( condition == true )
				sendString("SEARCH " + "SMALLER" + " " + sizeString);
			else
				sendString("SEARCH "+ "LARGER " + sizeString);
			
		}

		int i = 0;
		StringBuffer buffer = new StringBuffer();
		i = getAnswer();

		while (i != Imap4.ANSWER_OK)
		{
			if (i == Imap4.ANSWER_MORE)
			{
				buffer.append(answer + "\n");

				//cache.append( imap.answer+"\n");

				//Flags flags = changeFlags( cache.toString() );
			}
			else
			{

			}
			i = getAnswer();
		}

		System.out.println("buffer:\n" + buffer);
		
		result = Imap4Parser.parseSearch( buffer.toString() );

		return result;
	}
	
	protected String searchPattern( String pattern, boolean condition )
	{
		String searchpattern = new String();
		
		if ( pattern.equalsIgnoreCase("Answered") )
            {
                searchpattern = new String("ANSWERED");
            }
            else if ( pattern.equalsIgnoreCase("Deleted") )
            {
                searchpattern = new String("DELETE");
            }
            else if ( pattern.equalsIgnoreCase("Flagged") )
            {
                searchpattern = new String("FLAGGED");
            }
            else if ( pattern.equalsIgnoreCase("Recent") )
            {
                searchpattern = new String("RECENT");
            }
            else if ( pattern.equalsIgnoreCase("Draft") )
            {
                searchpattern = new String("DRAFT");
            }
            else if ( pattern.equalsIgnoreCase("Seen") )
            {
                searchpattern = new String("SEEN");
            }	
            
            if ( pattern.equalsIgnoreCase("RECENT") == false )
            {
            	if ( condition == false )
            	{
            		searchpattern = "UN"+searchpattern;
            	}
            }
            else
            {
            	if ( condition == false )
            		searchpattern = "OLD";
            }
           return searchpattern;
	}
	
	public Vector searchFlags(
		String messageSet,
		String pattern,
		boolean condition)
		throws Exception
	{
		Vector result = new Vector();

		String search = searchPattern( pattern, condition );
		
		if (messageSet.length() != 0)
		{
			if ( condition == true )
				sendString("SEARCH " + messageSet +" "+ search);
			else
				sendString("SEARCH " + messageSet + " " + search);
		}
		else
		{
			if ( condition == true )
				sendString("SEARCH" + " " + search);
			else
				sendString("SEARCH"+ " "+ search);
			
		}

		int i = 0;
		StringBuffer buffer = new StringBuffer();
		i = getAnswer();

		while (i != Imap4.ANSWER_OK)
		{
			if (i == Imap4.ANSWER_MORE)
			{
				buffer.append(answer + "\n");

				//cache.append( imap.answer+"\n");

				//Flags flags = changeFlags( cache.toString() );
			}
			else
			{

			}
			i = getAnswer();
		}

		System.out.println("buffer:\n" + buffer);
		
		result = Imap4Parser.parseSearch( buffer.toString() );

		return result;
	}
	
	
	
}
